import string
import secrets
from pydantic import TypeAdapter, ValidationError, ConfigDict
from typing import (
    Any,
    Literal,
    Callable,
    Type,
    TypeVar,
    overload,
    ParamSpec,
    Concatenate
)
from typing_extensions import TypedDict, NotRequired

alphabet = string.ascii_letters


def delete_none(dict_) -> dict:
    return {k: v for k, v in dict_.items() if v is not None}


def gen_op_id(length: int = 15) -> str:
    return ''.join(secrets.choice(alphabet) for i in range(length))


def validate_typeddict(
        typed_dict: type[TypedDict],
        data: dict[str, Any],
        extra: Literal['allow', 'ignore'] = 'forbid',
) -> bool:
    """
    Args:
        typed_dict: TypedDict Class
        data: dict
        extra: extra
    """
    class _ValidateClass(typed_dict): ...

    try:
        _ValidateClass.__pydantic_config__ = ConfigDict(extra=extra)
        TypeAdapter(_ValidateClass).validate_python(data, strict=True)
        return True
    except ValidationError as e:
        print(e)
        return False


def get_typeddict_validator(
        typed_dict: type[TypedDict],
        extra: Literal['allow', 'ignore'] = 'forbid',
        strict: bool = True,
) -> Callable[[dict[str, Any]], bool]:

    class _ValidateClass(typed_dict): pass

    def wrap(
            data: dict[str, Any],
    ) -> bool:
        try:
            _ValidateClass.__pydantic_config__ = ConfigDict(extra=extra)
            TypeAdapter(_ValidateClass).validate_python(data, strict=strict)
            return True
        except ValidationError as e:
            print(e)
            return False
    return wrap


P = ParamSpec("P")
TSelf = TypeVar("TSelf")
TReturn = TypeVar("TReturn")
T0 = TypeVar("T0")
T1 = TypeVar("T1")
T2 = TypeVar("T2")


@overload
def add_args_to_signature(
   to_signature: Callable[Concatenate[TSelf, P], TReturn],
   new_arg_type: Type[T0]
) -> Callable[
   [Callable[..., TReturn]],
   Callable[Concatenate[TSelf, T0, P], TReturn]
]:
   pass


@overload
def add_args_to_signature(
   to_signature: Callable[Concatenate[TSelf, P], TReturn],
   new_arg_type0: Type[T0],
   new_arg_type1: Type[T1],
) -> Callable[
   [Callable[..., TReturn]],
   Callable[Concatenate[TSelf, T0, T1, P], TReturn]
]:
   pass


@overload
def add_args_to_signature(
   to_signature: Callable[Concatenate[TSelf, P], TReturn],
   new_arg_type0: Type[T0],
   new_arg_type1: Type[T1],
   new_arg_type2: Type[T2]
) -> Callable[
   [Callable[..., TReturn]],
   Callable[Concatenate[TSelf, T0, T1, P], TReturn]
]:
   pass

# repeat if you want to enable adding more parameters...

def add_args_to_signature(
   *_, **__
):
   return lambda f: f



from fastapi import FastAPI
from fastapi_namespace import Resource
from fastapi_namespace import Namespace
import typing as t
import typing_extensions as tx

namespace = Namespace(prefix='', tags=[''])

class Test(tx.TypedDict):
    test: str
    a: int



app = FastAPI()
app.include_router(namespace)


if __name__ == "__main__":
    from uvicorn import run
    run(
        app=app,
        port=23457
    )


from fastapi.security.base import SecurityBase as SecurityBase_
from typing import ParamSpec, Callable, TypeVar, Optional

P = ParamSpec("P")
T = TypeVar("T")
ValidateFunction = Callable[P, T]


class SecurityBase(SecurityBase_):...
    # def __init__(self):
    #     self._validate: Optional[ValidateFunction] = None
    #
    # def validate(self, func: ValidateFunction) -> ValidateFunction:


from starlette.routing import BaseRoute
from fastapi import Depends, Response
from fastapi.types import IncEx
from fastapi.routing import APIRoute

from typing import Sequence, Type, Any, Callable, TypeVar
from typing_extensions import (
    TypedDict,
    NotRequired
)
from enum import Enum


class MethodDocument(TypedDict):
    response_model: NotRequired[Any]
    status_code: NotRequired[int]
    tags: NotRequired[list[str | Enum]]
    dependencies: NotRequired[Sequence[Depends]]
    summary: NotRequired[str]
    description: NotRequired[str]
    response_description: NotRequired[str]
    responses: NotRequired[dict[[int | str], dict[str, Any]]]
    deprecated: NotRequired[bool]
    operation_id: NotRequired[str]
    response_model_include: NotRequired[IncEx]
    response_model_exclude: NotRequired[IncEx]
    response_model_by_alias: NotRequired[bool]
    response_model_exclude_unset: NotRequired[bool]
    response_model_exclude_defaults: NotRequired[bool]
    response_model_exclude_none: NotRequired[bool]
    include_in_schema: NotRequired[bool]
    response_class: NotRequired[Type[Response]]
    name: NotRequired[str]
    callbacks: NotRequired[list[BaseRoute]]
    openapi_extra: NotRequired[dict[str, Any]]
    generate_unique_id_function: NotRequired[Callable[[APIRoute], str]]


DecoratedCallable = TypeVar("DecoratedCallable", bound=Callable[..., Any])


import uuid

from .typings import TokenPayload, UserIdentify, RawToken, TokenIdentify
from typing import TypeVar, Generic, Optional, Any, Mapping
from orjson import dumps, loads

T = TypeVar('T', bound="Token")


class _Base:
    def dumps(self, except_none: bool = True) -> bytes:
        return dumps(self.as_dict(except_none=except_none))

    def as_dict(self, except_none: bool = True) -> dict[str, Any]:
        dict_ = {}
        for k, v in self.__dict__.items():
            if isinstance(v, _Base):
                dict_[k] = v.as_dict(except_none=except_none)
            else:
                if except_none:
                    if v is not None:
                        dict_[k] = v
        return dict_

    def update(self, data: Mapping[str, Any]):
        for k, v in data.items():
            if (getattr(self, k, None)) is None:
                continue
            setattr(self, k, v)


class Token(_Base):
    def __init__(
            self,
            payload: TokenPayload,
            uid: UserIdentify
    ):
        self.payload = payload
        self.uid = uid


class TokenInfo(Generic[T], _Base):
    def __init__(
            self,
            info: T,
            token: RawToken,
            expires: int = 0,
    ):
        self.info = info
        self.token = token
        self.expires = expires


class OpaqueToken(Token):
    def __init__(
            self,
            payload: TokenPayload,
            uid: UserIdentify,
            idf: TokenIdentify = None
    ):
        super().__init__(
            payload,
            uid,
        )
        self.idf = idf


class OpaqueTokenInfo(TokenInfo[OpaqueToken]):
    pass







from .tokenBaseMixin import (
    TokenBaseMixin,
    TokenValidator,
    TokenInfoValidator,
    TI
)
from .typings import (
    OpaqueTokenInfo,
    OpaqueToken,
    JWTToken,
    TokenInfo,
    RawToken,
    Token
)
from fastapi_namespace.utils import get_typeddict_validator
from secrets import token_urlsafe
from uuid import uuid4
from typing_extensions import TypedDict
from dataclasses import asdict


class OpaqueTokenMixin(TokenBaseMixin[OpaqueToken, OpaqueTokenInfo]):

    def __init__(self):
        token_validator = get_typeddict_validator(OpaqueToken)
        token_info_validator = get_typeddict_validator(OpaqueTokenInfo)

        super().__init__(
            token=OpaqueToken,
            token_info=OpaqueTokenInfo,
            token_validator=token_validator,
            token_info_validator=token_info_validator
        )

    def create_token(self, *args, **kwarg) -> RawToken:
        """
        Returns:
            RawToken: `str`으로 된 토큰
        """
        return token_urlsafe(48)

    def _make_token(self, token: Token) -> OpaqueTokenInfo:
        # return OpaqueTokenInfo(
        #     **token,
        #     idf=uuid4().hex
        # )
        OpaqueToken(**asdict(token))
        return OpaqueTokenInfo(
            info=OpaqueToken(**asdict(token)),
        )




import uuid

from redis.asyncio import Redis
from redis.asyncio.client import Pipeline
from redis import Redis
AsyncRedis = Redis
"""
Async ``RedisClient``
"""

AsyncPipeline = Pipeline
"""
Async ``RedisClientPipeline``
"""

from typing import (
    Literal,
    NewType,
    Any,
    Union,
    Generic,
    Optional,
    TypeVar,
    TypeAlias
)
from typing_extensions import (
    TypedDict,
    NotRequired
)
from dataclasses import dataclass, field


TokenKey = NewType('TokenKey', str)
"""
토큰 키 **{user token prefix} / {token}** 형태
"""
UserTokenKey = NewType('UserTokenKey', str)
"""
유저 토큰 키 **{user token prefix} / {uid}** 형태
"""

MethodType = Literal["get", "post", "put", "delete", "patch", "head", "options", "trace"]

TokenType = Literal["ACCESS", "REFRESH"]

RawToken = str

TokenLimit = int

TokenExpire = int

TokenIdentify = str

TokenPayload = dict[str, Any]

UserIdentify = Union[int, str]



@dataclass
class Token:
    """토큰 Base

        다른 토큰 Mixin 생성시 이것을 상속 받을 것

        Attributes:
            payload:
                토큰 페이 로드
            uid:
                Redis 에 등록될 때 구분할 uid
        """
    payload: TokenPayload
    uid: UserIdentify


@dataclass
class OpaqueToken(Token):
    idf: TokenIdentify = field(default=lambda x: uuid.uuid4().hex)


@dataclass
class JWTToken(Token):
    jti: TokenIdentify = field(default=lambda x: uuid.uuid4().hex)


T = TypeVar('T', bound=Token)


@dataclass
class TokenInfo(Generic[T]):
    info: T
    token: Optional[RawToken] = ''
    expires_in: Optional[TokenExpire] = 0


@dataclass
class OpaqueTokenInfo(TokenInfo[OpaqueToken]):
    info: OpaqueToken


@dataclass
class JWTTokenInfo(TokenInfo[JWTToken]):
    info: JWTToken






from ..mixinBase import MixinBase
from .typings import (
    TokenLimit,
    TokenKey,
    UserTokenKey,
    TokenExpire,
    MethodType,
    OpaqueToken,
    OpaqueTokenInfo,
    JWTToken,
    JWTTokenInfo,
    TokenType,
    RawToken,
    UserIdentify,
    AsyncRedis,
    AsyncPipeline,
    TokenPayload,
    Token,
TokenInfo
)
from typing import (
    Callable,
    Union,
    Awaitable,
    ParamSpec,
    TypeAlias,
    Generic,
    TypeVar,
    Iterable,
    Coroutine,
    Any,
    Optional
)
from abc import abstractmethod
import asyncio
from functools import partial
from orjson import dumps as _dumps, loads
from .sercurity.base import SecurityBase
from typing import Union
from dataclasses import dataclass, asdict



T = TypeVar("T", bound=Token)
# TI = TypeVar("TI", OpaqueTokenInfo, JWTTokenInfo)
TI = TypeVar("TI", bound=TokenInfo)

TokenKeyHandler = Callable[[RawToken], TokenKey]
UserTokenKeyHandler = Callable[[UserIdentify], UserTokenKey]
_validator: TypeAlias = Callable[[dict[str, Any]], bool]
TokenValidator = _validator
TokenInfoValidator = _validator
# CallableToken = Callable[..., T]
# CallableTokenInfo = Callable[..., TI]
CallableToken = type[T]
CallableTokenInfo = type[TI]

def dumps(data):
    print(data, 'werwer')
    return _dumps(asdict(data))




class TransactionCore:
    @staticmethod
    async def _manage_user_token_transaction(pipe: AsyncPipeline, key: UserTokenKey):
        """
        ``Redis`` 내 토큰 관리

        만료 된 토큰은 ``UserTokenList`` 에서 삭제 함
        """
        if len((tokens := await pipe.smembers(key))) == 0:
            return
        await pipe.watch(*tokens)

        async def _gen():
            for i in tokens:
                yield i, await pipe.exists(i)

        tokens_for_delete = [i[0] async for i in _gen() if i[1] == 0]
        if len(tokens_for_delete) == 0:
            return
        pipe.multi()
        await pipe.srem(key, *tokens_for_delete)

    @staticmethod
    async def _manage_user_token_count_transaction(
            pipe: AsyncPipeline,
            key: UserTokenKey,
            limit: TokenLimit
    ) -> None:
        """``Redis`` 내에 토큰 갯수 관리
        """
        if len((tokens := await pipe.smembers(key))) == 0:
            return
        await pipe.watch(*tokens)
        token_count = len(tokens)

        if token_count >= limit:
            delete_count = limit - token_count + 1  # 무조건 음수
            delete_tokens = [(i, await pipe.ttl(i)) for i in tokens]
            delete_tokens.sort(key=lambda x: x[1])
            delete_tokens = delete_tokens[:delete_count]
            pipe.multi()
            _: Awaitable = pipe.delete(*(d := [i[0] for i in delete_tokens]))
            _: Awaitable = pipe.srem(key, *d)
        return

    @staticmethod
    async def _add_token_transaction(
            pipe: AsyncPipeline,
            user_token_key: UserTokenKey,
            token_key: TokenKey,
            token: Token,
            token_expire: TokenExpire
    ) -> bool:
        """``Redis`` 내에 토큰 등록
        """
        token = dumps(token)
        if await pipe.sismember(user_token_key, token_key) == 1:
            pipe.multi()
            _: Awaitable = pipe.set(token_key, token)
            _: Awaitable = pipe.expire(token_key, token_expire)
            return True
        else:
            return False

    @staticmethod
    async def _get_token_transaction(
            pipe: AsyncPipeline,
            raw_token: RawToken,
            key: TokenKey,
            token: CallableToken,
            token_info: CallableTokenInfo,
            token_validator: TokenValidator,
    ) -> Optional[TI]:
        """``Redis`` 에서 토큰 가져오기
        """
        if (__token := await pipe.get(key)) is None:
            await pipe.unwatch()
            return None

        _token = token(**loads(__token))
        if not token_validator(_token):
            return None

        ttl = await pipe.ttl(key)

        return token_info(
            token=raw_token,
            expires_in=ttl,
            **_token
        )

    @staticmethod
    async def _get_user_tokens_transaction(
            pipe: AsyncPipeline,
            key: UserTokenKey,
            token: CallableToken,
            token_info: CallableTokenInfo,
            token_validator: TokenValidator,
    ) -> list[TI]:
        """``Redis`` 에서 특정 유저의 토큰 가져오기
        """
        tokens = await pipe.smembers(key)
        if len(tokens) == 0:
            return []

        await pipe.watch(*tokens)
        ttl = None
        return [
            token_info(
                token=i.split('/')[1],
                expires_in=ttl,
                **_token
            )
            for i in tokens if token_validator((_token := token(**loads(await pipe.get(i))))) and (ttl := await pipe.ttl(i)) > 5
        ]

    @staticmethod
    async def _abort_user_token_transaction(
            pipe: AsyncPipeline,
            key: UserTokenKey,
            get_token_key: TokenKeyHandler,
            user_tokens: list[RawToken] | None = None
    ) -> None:
        """``Redis`` 에서 특정 유저의 토큰을 취소
        """
        tokens = await pipe.smembers(key)
        if len(tokens) == 0:
            return

        if user_tokens is None or len(user_tokens) == 0:
            tokens_for_delete = list(tokens)
        else:
            tokens_for_delete = [d for i in user_tokens if (d := get_token_key(i)) in tokens]

        if len(tokens_for_delete) == 0:
            return

        pipe.multi()
        _: Awaitable = pipe.srem(key, *tokens_for_delete)
        _: Awaitable = pipe.delete(*tokens_for_delete)





class TokenBaseMixin(Generic[T, TI], MixinBase, TransactionCore):
    access_token_limit: Optional[TokenLimit] | None = 1
    refresh_token_limit: Optional[TokenLimit] | None = 1
    access_token_expire: Optional[TokenExpire] = 3600
    refresh_token_expire: Optional[TokenExpire] = 129600
    user_access_token_key: Optional[str] = 'user_access_token'
    user_refresh_token_key: Optional[str] = 'user_refresh_token'
    access_token_key: Optional[str] = 'access_token'
    refresh_token_key: Optional[str] = 'refresh_token'
    security_route: Optional[list[MethodType]] = []
    security_class: Optional[SecurityBase] = None

    def __init__(
            self,
            token: type[T],
            token_info: type[TI],
            token_validator: TokenValidator,
            token_info_validator: TokenInfoValidator
    ):
        self._Token = token
        self._TokenInfo = token_info
        self._token_validator = token_validator
        self._token_info_validator = token_info_validator

    @abstractmethod
    def create_token(self, *args, **kwarg) -> RawToken:
        pass

    @abstractmethod
    def _make_token(self, token: Token) -> TI:
        pass

    def _get_access_token_key(self, token: RawToken) -> TokenKey:
        return TokenKey(f"{self.access_token_key}/{token}")

    def _get_refresh_token_key(self, token: RawToken) -> TokenKey:
        return TokenKey(f"{self.refresh_token_key}/{token}")

    def _get_token_key_handler(self, type: TokenType) -> TokenKeyHandler:
        if type == "ACCESS":
            return self._get_access_token_key
        elif type == "REFRESH":
            return self._get_refresh_token_key
        raise ValueError(f"Token type must be ACCESS or REFRESH")

    def _get_user_access_token_key(self, idf: UserIdentify) -> UserTokenKey:
        return UserTokenKey(f"{self.user_access_token_key}/{idf}")

    def _get_user_refresh_token_key(self, idf: UserIdentify) -> UserTokenKey:
        return UserTokenKey(f"{self.user_refresh_token_key}/{idf}")

    def _get_user_token_key_handler(self, type: TokenType) -> UserTokenKeyHandler:
        if type == "ACCESS":
            return self._get_user_access_token_key
        elif type == "REFRESH":
            return self._get_user_refresh_token_key
        raise ValueError(f"Token type must be ACCESS or REFRESH")

    async def _register_token(
            self,
            rd: AsyncRedis,
            key: UserTokenKey,
            token_key_handler: TokenKeyHandler,
            token_expire: TokenExpire,
            **kwargs
    ) -> tuple[RawToken, TokenKey]:
        """``Redis`` 에 토큰 등록

        * class 내부 사용됨

        Args:
            rd: Async Redis Client
            key:
            token_key_handler:
            token_expire: 토큰 만료 시간
            **kwargs: 토큰 만들때 인자로 넘겨줌

        Returns:
            * ``tuple[0]`` = ``RawToken``
            * ``tuple[1]`` = ``TokenKey``
        """

        while await rd.sadd(
                key,
                (
                        token_key := token_key_handler(
                            token := self.create_token(**kwargs)
                        )
                )
        ) == 0:
            pass
        await rd.expire(key, token_expire)
        return token, TokenKey(token_key)

    async def _manage_user_token(
            self,
            rd: AsyncRedis,
            key: UserTokenKey
    ) -> None:
        """uid로 등록된 토큰을 관리함

        * class 내부 사용

        Args:
            rd: Async Redis client
            key:
        """
        await rd.transaction(
            partial(self._manage_user_token_transaction, key=key),
            key
        )

    async def _manage_user_token_count(
            self,
            rd: AsyncRedis,
            key: UserTokenKey,
            limit: TokenLimit
    ) -> None:
        """토큰 갯수 제한시 토큰을 관리함

        * class 내부 사용

        Args:
            rd: Async Redis client
            key: User Token key
            limit: 토큰 제한 갯수
        """
        await rd.transaction(
            partial(self._manage_user_token_count_transaction, key=key, limit=limit),
            key
        )

    async def _create_type_token(
            self,
            rd: AsyncRedis,
            payload: TokenPayload,
            identify: UserIdentify,
            type: TokenType,
            **kwargs,
    ) -> TI:
        """토큰 만든 후 Redis에 저장

        * Class 내부 사용
        * 직접 사용 금지 ``create_access_token`` ``create_refresh_token`` 을 사용

        Args:
            rd: Async Redis Client
            payload:
            identify: Redis 내 사용될 uid
            type: ACCESS or REFRESH
            **kwargs: create_token 의 인자로 넘겨줌
        """
        if type == 'ACCESS':
            token_limit = self.access_token_limit
            token_expire = self.access_token_expire
        elif type == 'REFRESH':
            token_limit = self.refresh_token_limit
            token_expire = self.refresh_token_expire
        else:
            raise AttributeError("type must be ACCESS or REFRESH")

        get_token_key = self._get_token_key_handler(type)
        get_user_token_key = self._get_user_token_key_handler(type)

        user_token_key = get_user_token_key(identify)

        # 토큰 갯수 및 존재 여부 관리
        await self._manage_user_token(rd=rd, key=user_token_key)
        if token_limit is not None:
            await self._manage_user_token_count(rd=rd, key=user_token_key, limit=token_limit)

        # 토큰 제작
        get_token_info = partial[TI](self._make_token, token=Token(uid=identify, payload=payload))
        get_raw_token_and_token_key = partial[Coroutine[Any, Any, tuple[RawToken, TokenKey]]](
            self._register_token,
            rd=rd,
            key=user_token_key,
            token_key_handler=get_token_key,
            token_expire=token_expire,
            **kwargs
        )
        while not await rd.transaction(
                partial(
                    self._add_token_transaction,
                    user_token_key=user_token_key,
                    token_key=(result := await get_raw_token_and_token_key())[1],
                    token=(token_info := get_token_info()),
                    token_expire=token_expire
                ),
                *(user_token_key, result[1]),
                value_from_callable=True
        ):
            pass

        ttl: int = await rd.ttl(result[1])
        token_info["expires_in"] = ttl
        token_info["token"] = result[0]
        return token_info

    async def create_access_token(
            self,
            rd: AsyncRedis,
            payload: TokenPayload,
            identify: UserIdentify,
            **kwargs
    ) -> TI:
        """
        Args:
            rd:
            payload:
            identify:
            **kwargs: ``create_token`` 인자로 넘어감
        """
        return await self._create_type_token(
            rd=rd,
            payload=payload,
            type="ACCESS",
            identify=identify,
            **kwargs
        )

    async def create_refresh_token(
            self,
            rd: AsyncRedis,
            payload: TokenPayload,
            identify: UserIdentify,
            **kwargs
    ) -> TI:
        """
        Args:
            rd:
            payload:
            identify:
            **kwargs: ``create_token`` 인자로 넘어감
        """
        return await self._create_type_token(
            rd=rd,
            payload=payload,
            type="REFRESH",
            identify=identify,
            **kwargs
        )

    async def get_type_token(
            self,
            rd: AsyncRedis,
            token: RawToken,
            type: Optional[TokenType] = None,
    ) -> Optional[TI | tuple[TI, TI]]:
        """
        Returns:
            type == None 이면

            * tuple[0] : ACCESS TOKEN INFO
            * tuple[1] : REFRESH TOKEN INFO
        """
        kwargs = {
            "token": self._Token,
            "token_info": self._TokenInfo,
            "token_validator": self._token_validator,
        }
        if type is None:
            return await asyncio.gather(
                rd.transaction(
                    partial(
                        self._get_token_transaction,
                        raw_token=token,
                        key=(token_key := self._get_token_key_handler("ACCESS")(token)),
                        **kwargs
                    ),
                    token_key,
                    value_from_callable=True
                ),
                rd.transaction(
                    partial(
                        self._get_token_transaction,
                        raw_token=token,
                        key=(token_key := self._get_token_key_handler("REFRESH")(token)),
                        **kwargs
                    ),
                    token_key,
                    value_from_callable=True
                ),
            )
        else:
            get_token_key = self._get_token_key_handler(type)
            token_key = get_token_key(token)
            return await rd.transaction(
                partial(
                    self._get_token_transaction,
                    raw_token=token,
                    key=token_key,
                    **kwargs
                ),
                token_key,
                value_from_callable=True
            )

    async def get_access_token(
            self,
            rd: AsyncRedis,
            token: RawToken,
    ) -> Optional[TI]:
        return await self.get_type_token(rd=rd, token=token, type="ACCESS")

    async def get_refresh_token(
            self,
            rd: AsyncRedis,
            token: RawToken,
    ) -> Optional[TI]:
        return await self.get_type_token(rd=rd, token=token, type="REFRESH")

    async def _get_user_type_tokens(
            self,
            rd: AsyncRedis,
            identify: UserIdentify,
            type: TokenType = "ACCESS",
    ) -> list[TI]:
        get_token_key = self._get_token_key_handler(type)
        get_user_token_key = self._get_user_token_key_handler(type)

        key = get_user_token_key(identify)

        await self._manage_user_token(rd=rd, key=key)
        res: list[TI] = await rd.transaction(
            partial(
                self._get_user_tokens_transaction,
                key=key,
                token=self._Token,
                token_info=self._TokenInfo,
                token_validator=self._token_validator,
            ),
            key,
            value_from_callable=True
        )
        res.sort(key=lambda t: t['expires_in'])

        return res

    async def get_user_access_tokens(
            self,
            rd: AsyncRedis,
            identify: UserIdentify
    ) -> list[TI]:
        return await self._get_user_type_tokens(
            rd=rd,
            identify=identify,
            type="ACCESS"
        )

    async def get_user_refresh_tokens(
            self,
            rd: AsyncRedis,
            identify: UserIdentify
    ) -> list[TI]:
        return await self._get_user_type_tokens(
            rd=rd,
            identify=identify,
            type="REFRESH"
        )

    async def _abort_user_type_token(
            self,
            rd: AsyncRedis,
            identify: UserIdentify,
            type: TokenType,
            user_tokens: list[RawToken] | None = None,
    ) -> None:
        get_token_key = self._get_token_key_handler(type)
        get_user_token_key = self._get_user_token_key_handler(type)
        user_token_key = get_user_token_key(identify)

        await rd.transaction(
            partial(self._abort_user_token_transaction, key=user_token_key, get_token_key=get_token_key, user_tokens=user_tokens),
            user_token_key
        )

    async def abort_user_access_token(
            self,
            rd: AsyncRedis,
            identify: UserIdentify,
            user_tokens: RawToken | Iterable[RawToken] | None = None
    ) -> None:
        default = partial(self._abort_user_type_token, rd=rd, identify=identify, type="ACCESS")
        if user_tokens is None:
            await default()
        elif isinstance(user_tokens, Iterable):
            await default(user_tokens=[i for i in user_tokens])
        elif isinstance(user_tokens, str):
            await default(user_token=[user_tokens])
        else:
            raise ValueError('user Token error')

    async def abort_user_refresh_token(
            self,
            rd: AsyncRedis,
            identify: UserIdentify,
            user_tokens: RawToken | Iterable[RawToken] | None = None
    ) -> None:
        default = partial(self._abort_user_type_token, rd=rd, identify=identify, type="REFRESH")
        if user_tokens is None:
            await default()
        elif isinstance(user_tokens, Iterable):
            await default(user_tokens=[i for i in user_tokens])
        elif isinstance(user_tokens, str):
            await default(user_token=[user_tokens])
        else:
            raise ValueError('user Token error')

    async def abort_token(
            self,
            rd: AsyncRedis,
            token: RawToken,
    ) -> None:
        get_access_token_key = self._get_token_key_handler("ACCESS")
        get_refresh_token_key = self._get_token_key_handler("REFRESH")

        await asyncio.gather(
            rd.delete(get_access_token_key(token)),
            rd.delete(get_refresh_token_key(token))
        )

    async def _abort_type_token(
            self,
            rd: AsyncRedis,
            token: RawToken,
            type: TokenType
    ) -> None:
        token_key = self._get_token_key_handler(type)(token)
        await rd.delete(token_key)

    async def abort_access_token(
            self,
            rd: AsyncRedis,
            token: RawToken,
    ) -> None:
        await self._abort_type_token(rd=rd, token=token, type='ACCESS')

    async def abort_refresh_token(
            self,
            rd: AsyncRedis,
            token: RawToken,
    ) -> None:
        await self._abort_type_token(rd=rd, token=token, type='REFRESH')


from fastapi.security.base import SecurityBase as SecurityBase_
from typing import ParamSpec, Callable, TypeVar, Optional, Generic
from dataclasses import dataclass

P = ParamSpec("P")
TBase = TypeVar("TBase")
ValidateFunction = Callable[P, TBase]
T = TypeVar("T", )


class SecurityBase:
    def __init__(self):
        self.validate: Optional[Callable] = None


def validate(a: str):
    print('hi')


sec = SecurityBase()
sec.validate = validate
sec.validate()



def catch_exception(function: Callable[P, T]) -> Callable[P, Optional[T]]:
    def decorator(*args: P.args, **kwargs: P.kwargs) -> Optional[T]:
        try:
            return function(*args, **kwargs)
        except Exception:
            return None
    return decorator

@catch_exception
def div(arg: int) -> float:
    return arg / arg

a = div()



from .base import SecurityBase
from fastapi.security import (
    OAuth2 as OAuth2_,
    OAuth2AuthorizationCodeBearer as OAuth2AuthorizationCodeBearer_
)
from fastapi.openapi.models import OAuthFlows as OAuthFlowsModel
from fastapi.exceptions import HTTPException
from starlette.requests import Request
from starlette.status import HTTP_401_UNAUTHORIZED, HTTP_403_FORBIDDEN

from typing import (
    Union,
    Any,
    Optional,
    Callable,
    TypeVar
)


class OAuth2(SecurityBase, OAuth2_):
    def __init__(
            self,
            flows: Union[OAuthFlowsModel, dict[str, dict[str, Any]]],
            scheme_name: Optional[str] = None,
            description: Optional[str] = None,
            auto_error: bool = True,
    ):
        super().__init__(
            flows=flows,
            scheme_name=scheme_name,
            description=description,
            auto_error=auto_error
        )

    async def __call__(self, request: Request) -> Optional[str]:
        authorization = request.headers.get("Authorization")
        if not authorization:
            if self.auto_error:
                raise HTTPException(
                    status_code=HTTP_403_FORBIDDEN, detail="Not authenticated"
                )
            else:
                return None
        return authorization